/*
 * T030 ROM BIOS - Command interpreter
 *
 * (c) 2016 Tobias Rathje <tobbe@fasanix.com>
 *
 */

#include <stdio.h>
#include "mfp.h"
#include "xfer.h"
#include "vector.h"
#include "misc.h"
#include "netload.h"
#include "ram_test.h"
#include "lxload.h"
#include "sd.h"
#include "ehb.h"
#include "script.h"

char cmd_buf[128];

#define DEFAULT_ADDR 0x1000000

#define TEST_BIOS_ADDR 0x90000
#define TEST_BIOS_FILE "t030test.bin"

/*
 * Print prompt
 */
void prompt()
{
	mfp_puts("\r\n#> ");
}

/*
 * Print OK
 */
void send_ok()
{
	mfp_printf("OK\r\n");
}

/*
 * Unknown command
 */
void cmd_unknown()
{
	mfp_printf("Unknown command - h for help\r\n");
}

/*
 * Receive file with XMODEM
 */
void cmd_xfer(char *s)
{
	uint32_t addr;

	siscanf(s, "%"SCNx32, &addr);
	xfer_rcv(addr);
}

/*
 * Dump memory content
 */
void cmd_dump(char *s)
{
	uint32_t addr;
	uint8_t *data;
	uint8_t darr[16];

	siscanf(s, "%"SCNx32, &addr);

	for (int i = 0; i < 8; i++)
	{
		mfp_printf("%08X: ", addr);

		for (int j = 0; j < 16; j++)
		{
			data = (uint8_t *)addr++;
			mfp_printf("%02X ", *data);
			darr[j] = *data;
		}

		for (int j = 0; j < 16; j++)
		{
			// unprintable chars
			if (darr[j] < 32 || darr[j] > 126)
				mfp_printf(".");
			else
				mfp_printf("%c", darr[j]);
		}

		mfp_printf("\r\n");
	}
}

/*
 * Write byte to memory location
 */
void cmd_write(char *s)
{
	uint8_t data, cdata;
	volatile uint8_t *dp;
	uint32_t addr;
	unsigned int dt;

	siscanf(s, "%"SCNx32" %02x", &addr, &dt);

	data = dt;
	dp = (uint8_t *)addr;
	*dp = data;
	cdata = *dp;

	if (data == cdata)
		send_ok();
	else
		mfp_printf("ERROR\r\n");
}

/*
 * Download file from TFTP server to memory
 */
void cmd_netload(char *s)
{
	char lcl_ip[15];
	char rmt_ip[15];
	char fname[60];
	uint32_t addr;

	int ret = siscanf(s, "%s %s %s %"SCNx32, lcl_ip, rmt_ip, fname, &addr);

	if (ret > 0)
		do_netload(lcl_ip, rmt_ip, fname, addr);
	else
		do_netload(LCL_IP_DHCP, DEFAULT_TFTP_SERVER, "image.bin", DEFAULT_ADDR);
}

/*
 * Load Linux kernel image from TFTP server to memory
 */
void cmd_netload_kernel()
{
	netload_kernel();
}

/*
 * Load and execute Linux kernel
 */
void cmd_load_kernel(char *s)
{
	char cmd_line[80];
	uint32_t rd_addr, rd_size;

	int ret = siscanf(s, " \"%[^\"]\" %"SCNx32" %"SCNx32, cmd_line, &rd_addr, &rd_size);

	if (ret > 0)
		load_kernel(cmd_line, rd_addr, rd_size);
	else
		load_kernel("", 0, 0);
}

/*
 * Force various exceptions
 */
void cmd_test_exception(char *s)
{
	int n;
	siscanf(s, "%d", &n);
	test_exception(n);
}

/*
 * Execute from address
 */
void cmd_run(char *s)
{
	uint32_t addr;

	siscanf(s, "%"SCNx32, &addr);

	__asm("move.l %0,%%a0":: "d"(addr));
	__asm("jsr @(%a0)");
}

/*
 * Execute from address with user privilege
 */
void cmd_run_user(char *s)
{
	static uint32_t addr;

	siscanf(s, "%"SCNx32, &addr);

	// switch to user mode
	__asm("move.w #0x0100, %sr");

	// set up user stack
	__asm("move.l #0x000efffc, %sp");

	__asm("move.l %0,%%a0":: "d"(addr));
	__asm("jsr @(%a0)");

	// return to supervisor mode
	__asm("trap #14");
}

/*
 * Print content of CPU registers
 */
void cmd_dump_reg()
{
	static uint16_t sr;
	static uint32_t ssp, usp, vbr, cacr;

	__asm("move.w %%sr,%0":"=d"(sr));
	__asm("move.l %%ssp,%0":"=d"(ssp));
	__asm("move.l %%usp,%0":"=a"(usp));
	__asm("movec %%vbr,%0":"=d"(vbr));
	__asm("movec %%cacr,%0":"=d"(cacr));

	mfp_printf("SR: %04x\r\n", sr);
	mfp_printf("SSP: %08x\r\n", ssp);
	mfp_printf("USP: %08x\r\n", usp);
	mfp_printf("VBR: %08x\r\n", vbr);
	mfp_printf("CACR: %08x\r\n", cacr);
}

/*
 * Modify cache control register
 */
void cmd_cache_ctl(char *s)
{
	static uint32_t val;

	siscanf(s, "%"SCNx32, &val);
	cache_ctl(val);
}

/*
 * RAM test
 */
void cmd_check_ram(char *s)
{
	uint32_t addr, size;
	int m, c;

	siscanf(s, "%d %"SCNx32" %"SCNx32" %d", &m, &addr, &size, &c);

	switch (m)
	{
		case 0:
			check_base_ram();
			break;
		case 1:
			check_ext_ram();
			break;
		case 8:
			check_ram_8(addr, size, c);
			break;
		case 16:
			check_ram_16(addr, size, c);
			break;
		case 32:
			check_ram_32(addr, size, c);
			break;
		default:
			mfp_printf("Unknown check mode\r\n");
	}
}

/*
 * Change interrupt level
 */
void cmd_int_level(char *s)
{
	int i;
	static uint16_t l;

	siscanf(s, "%d", &i);

	l = i;
	l = l << 8;
	l |= 0x2000;

	__asm("move %0,%%sr":: "d"(l));
}

/*
 * Toggle heartbeat LED
 */
void cmd_hb(char *s)
{
	int x;

	siscanf(s, "%d", &x);

	if (x)
		enable_hb_led();
	else
		disable_hb_led();
}

/*
 * Print ASCII mandelbrot
 */
void cmd_mandel(char *s)
{
	int n;
	siscanf(s, "%d", &n);

	for (int i = 0; i < n; i++)
	{
		mfp_printf("%d:\r\n", i);
		print_mandel();
	}
}

/*
 * Load and execute test monitor from TFTP server
 */
void cmd_test_bios_tftp()
{
	do_netload(LCL_IP_DHCP, DEFAULT_TFTP_SERVER, TEST_BIOS_FILE, TEST_BIOS_ADDR);

	__asm("move.l %0,%%a0":: "d"(TEST_BIOS_ADDR));
	__asm("jsr @(%a0)");
}

/*
 * Load and execute test monitor from SD card
 */
void cmd_test_bios_sd()
{
	sd_read_file(TEST_BIOS_FILE, TEST_BIOS_ADDR);

	__asm("move.l %0,%%a0":: "d"(TEST_BIOS_ADDR));
	__asm("jsr @(%a0)");
}

/*
 * Soft reset (jump to reset vector)
 */
void cmd_soft_reset()
{
	uint32_t addr = (*(uint32_t *)0x4);

	cli();

	__asm("move.l %0,%%a0":: "d"(addr));
	__asm("jsr @(%a0)");
}

/*
 * Print help
 */
void cmd_help()
{
	mfp_printf("a  - Load and execute EhBasic\r\n");
	mfp_printf("b  - Toggle heartbeat LED: b <val>\r\n");
	mfp_printf("c  - Modify cache register: c <val>\r\n");
	mfp_printf("d  - Dump address: d <addr>\r\n");
	mfp_printf("e  - Force exception: e <num>\r\n");
	mfp_printf("h  - Help\r\n");
	mfp_printf("i  - Set interrupt level: i <level>\r\n");
	mfp_printf("g  - Run: g <addr>\r\n");
	mfp_printf("k  - Load linux kernel image from TFTP server\r\n");
	mfp_printf("l  - Load linux: l <cmdline> <ramdisk start addr> <ramdisk size>\r\n");
	mfp_printf("m  - Mandelbrot test: m <count>\r\n");
	mfp_printf("n  - Load image from TFTP server: n <local ip> <remote ip> <filename> <address>\r\n");
	mfp_printf("q  - Perform RAM check: q <mode> <address> <size> <count>\r\n");
	mfp_printf("r  - Dump registers\r\n");
	mfp_printf("sd - Delete file on SD card: sd <filename>\r\n");
	mfp_printf("sl - List files on SD card\r\n");
	mfp_printf("sn - Load image from TFTP server and store on SD card: sn <filename>\r\n");
	mfp_printf("sr - Read file from SD card to memory: sr <filename> <target addr>\r\n");
	mfp_printf("sw - Write file on SD card from memory: sw <filename> <source addr> <length>\r\n");
	mfp_printf("t  - Load and execute Test BIOS from TFTP server\r\n");
	mfp_printf("v  - Load and execute Test BIOS from SD card\r\n");
	mfp_printf("u  - Run in user mode: u <addr>\r\n");
	mfp_printf("w  - Write single byte: w <addr> <data>\r\n");
	mfp_printf("x  - Transfer file with XMODEM: x <addr>\r\n");
	mfp_printf("z  - Soft reset\r\n");
}

/*
 * Command interpreter
 */
void run_command(char *cmd_str)
{
	char *bp = cmd_str;
	char cmd = *bp++;

	switch (cmd)
	{
		case 'a':
			run_basic();
			break;

		case 'b':
			cmd_hb(bp);
			break;

		case 'c':
			cmd_cache_ctl(bp);
			break;

		case 'd':
			cmd_dump(bp);
			break;

		case 'e':
			cmd_test_exception(bp);
			break;

		case 'h':
			cmd_help();
			break;

		case 'i':
			cmd_int_level(bp);
			break;

		case 'g':
			cmd_run(bp);
			break;

		case 'k':
			cmd_netload_kernel();
			break;

		case 'l':
			cmd_load_kernel(bp);
			break;

		case 'm':
			cmd_mandel(bp);
			break;

		case 'n':
			cmd_netload(bp);
			break;

		case 'q':
			cmd_check_ram(bp);
			break;

		case 'r':
			cmd_dump_reg();
			break;

		case 's':
			sd_cmd(bp);
			break;

		case 't':
			cmd_test_bios_tftp();
			break;

		case 'v':
			cmd_test_bios_sd();
			break;

		case 'u':
			cmd_run_user(bp);
			break;

		case 'w':
			cmd_write(bp);
			break;

		case 'x':
			cmd_xfer(bp);
			break;

		case 'z':
			cmd_soft_reset();
			break;

		case '0':
			run_script(0);
			break;

		case '1':
			run_script(1);
			break;

		case '2':
			run_script(2);
			break;

		case '3':
			run_script(3);
			break;

		case '4':
			run_script(4);
			break;

		case '5':
			run_script(5);
			break;

		case '6':
			run_script(6);
			break;

		case '7':
			run_script(7);
			break;

		case '8':
			run_script(8);
			break;

		case '9':
			run_script(9);
			break;

		default:
			cmd_unknown();
	}
}

/*
 * Main command loop
 */
void cmd_loop()
{
	while(1)
	{
		prompt();

		int i = 0;

		while(1)
		{
			char c = mfp_getc();

			// end of input
			if (c == '\r')
			{
				cmd_buf[i] = '\0';
				break;
			}
			// backspace
			else if (c == 0x08)
			{
				i--;
				mfp_putc(c);
			}
			// break
			else if (c == 0x03)
			{
				i = 0;
				break;
			}
			// ignore newline
			else if (c == '\n')
			{
			}
			else
			{
				cmd_buf[i++] = c;
				mfp_putc(c);
			}
		}

		mfp_printf("\r\n");

		// empty cmd
		if (i == 0)
			continue;

		run_command(cmd_buf);
	}
}
