/*
 * T030 ROM BIOS - Linux loader
 *
 * (c) 2016 Tobias Rathje <tobbe@fasanix.com>
 *
 */

#include <stdio.h>
#include <string.h>

#include "mfp.h"
#include "netload.h"
#include "lxload.h"
#include "tftp.h"
#include "misc.h"

#define KERNEL_IMAGE "linux.bin"
#define KERNEL_ADDR 0x1001000

#define T030_MACHTYPE 13
#define T030_RAMSTART 0x1000000
#define T030_RAMSIZE 16*1024*1024

uint8_t *bi_ptr;

void netload_kernel()
{
	do_netload(LCL_IP_DHCP, DEFAULT_TFTP_SERVER, KERNEL_IMAGE, KERNEL_ADDR);
}

uint32_t get_bi_addr()
{
	uint32_t i;

	// get kernel size from TFTP loader
	i = last_xfer_size;

	// add RAM offset
	i = i + T030_RAMSTART;

	// add one page offset
	i = i + 4096;

	return i + (4096 - (i % 4096));
}

void add_bi_record(uint16_t tag, uint16_t size, uint32_t data)
{
	static struct bi_record bi;

	bi.tag = tag;
	bi.size = size;
	bi.data[0] = data;

	memcpy(bi_ptr, &bi, bi.size);
	bi_ptr = bi_ptr + bi.size;
}

void load_kernel(char *cmd_line, uint32_t rd_addr, uint32_t rd_size)
{
	uint32_t start_addr, bi_addr;
	struct mem_info mi, rd;

	// derive bi addr from kernel size
	bi_addr = get_bi_addr();

	mfp_printf("\r\nBootinfo address: 0x%08x\r\n", bi_addr);

	bi_ptr = (uint8_t *)bi_addr;

	add_bi_record(BI_MACHTYPE, 8, T030_MACHTYPE);
	add_bi_record(BI_CPUTYPE, 8, 2);
	add_bi_record(BI_FPUTYPE, 8, 0);
	add_bi_record(BI_MMUTYPE, 8, 2);

	// MEMCHUNK record
	*bi_ptr++ = 0;
	*bi_ptr++ = BI_MEMCHUNK;
	*bi_ptr++ = 0;
	*bi_ptr++ = 12; // record size
	mi.addr = T030_RAMSTART;
	mi.size = T030_RAMSIZE;
	memcpy(bi_ptr, &mi, 8);
	bi_ptr = bi_ptr + 8;

	// ramdisk
	if (rd_addr > 0)
	{
		*bi_ptr++ = 0;
		*bi_ptr++ = BI_RAMDISK;
		*bi_ptr++ = 0;
		*bi_ptr++ = 12; // record size
		rd.addr = rd_addr;
		rd.size = rd_size;
		memcpy(bi_ptr, &rd, 8);
		bi_ptr = bi_ptr + 8;

		mfp_printf("\r\nRamdisk start: 0x%08x\r\nRamdisk size: %d\r\n", rd.addr, rd.size);
	}

	// kernel cmd line
	*bi_ptr++ = 0;
	*bi_ptr++ = BI_COMMAND_LINE;
	*bi_ptr++ = 0;
	*bi_ptr++ = 4 + strlen(cmd_line);
	memcpy(bi_ptr, cmd_line, strlen(cmd_line));
	bi_ptr = bi_ptr + strlen(cmd_line);

	add_bi_record(BI_LAST, 8, 0);

	mfp_printf("\r\nMachine type: %d\r\nMemory start: 0x%08x\r\nMemory size: %d MB\r\n\r\n", T030_MACHTYPE, mi.addr, mi.size / (1024*1024));

	start_addr = KERNEL_ADDR;

	mfp_printf("Loading Linux at 0x%08x\r\n", start_addr);

	// disable interrupts
	cli();

	// disable cache
	cache_ctl(0);

	__asm("move.l %0,%%a0":: "d"(start_addr));
	__asm("jsr @(%a0)");
}

