/*
 * T030 ROM BIOS - Multifunction peripheral controller routines
 *
 * (c) 2016 Tobias Rathje <tobbe@fasanix.com>
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include "mfp.h"
#include "util.h"
#include "buffer.h"
#include "vector.h"

char pbuf[256];

strBuffer rxbuffer;

/*
 * Initialize MFP
 */
void mfp_init()
{
	MFPDDR = 0x3f;		// 6+7 inputs, rest outputs

	MFPTCDR = 0x02; 	// 1/4 tx clock
	MFPTDDR = 0x02;		// 1/4 rx clock
	MFPTCDCR = 0x11;	// divide by 4 in C/D control register
	MFPUCR = 0x88;		// divide by 16, 8 bit, no parity

	MFPVR = 0x40;		// MFP interrupt vectors start at 64

	MFPRSR = 0x01;		// start receiver clock
	MFPTSR = 0x05;		// start transmitter clock

	// init rx buffer
	bufferInitialise(&rxbuffer);
}

/*
 * Put char to USART
 */
void mfp_putc(unsigned char c)
{
	while (bit_is_clear(MFPTSR, 7)); // wait for empty transmit buffer
	MFPUDR = c; // transfer c
}

/*
 * Get char from USART
 */
int mfp_getc()
{
	while (bit_is_clear(MFPRSR, 7)); // wait for char in receiver buffer
	return MFPUDR;
}

/*
 * Get char from USART (non blocking, buffered)
 */
int mfp_getc_buf()
{
	uint8_t c;

	uint8_t res = bufferGet(&c, &rxbuffer);

	if (res)
		return c;
	else
		return -1;
}

/*
 * ISR for UART interrupt
 */
void __attribute__ ((interrupt)) mfp_uart_rx_isr()
{
	bufferPut(MFPUDR, &rxbuffer);
}

/*
 * Put string to USART
 */
void mfp_puts(char *s)
{
	while (*s != '\0')
		mfp_putc(*s++);
}

/*
 * Enable/disable UART rx interrupt
 */
void mfp_set_uart_int(int x)
{
	if (x > 0)
	{
		set_vector(76, mfp_uart_rx_isr);
		bit_set(MFPIERA, 4);
		bit_set(MFPIMRA, 4);
	}
	else
	{
		bit_clear(MFPIERA, 4);
	}
}

/*
 * printf to USART
 */
void mfp_printf(const char *format, ...)
{
	va_list args;
	va_start(args, format);
	vsiprintf(pbuf, format, args);
	mfp_puts(pbuf);
}

/*
 * Generate msec delay using MFP timer B (not precise at all!)
 */
void mfp_delay_ms(int d)
{
	for (int i = 0; i < d; i++)
	{
		MFPTBCR = 0;	// disable timer B
		MFPTBDR = 15;	// ~1 ms
		MFPTBCR = 7;	// start timer B, divide by 200

		while (MFPTBDR > 1); // wait until timer B reaches zero
	}
}
