/*
 * T030 ROM BIOS - Helper routines
 *
 * (c) 2016 Tobias Rathje <tobbe@fasanix.com>
 *
 */

#include <stdio.h>
#include "mfp.h"
#include "vector.h"

/*
 * Enable interrupts
 */
void sei()
{
	__asm("move.w #0x2100, %sr");
}

/*
 * Disable interrupts
 */
void cli()
{
	__asm("move.w #0x2700, %sr");
}

/*
 * Control cache register
 */
void cache_ctl(uint32_t val)
{
	__asm("movec %0,%%cacr":: "d"(val));
}

/*
 * LED heartbeat ISR
 */
void __attribute__ ((interrupt)) timerb_isr()
{
	static uint8_t c;

	if (c++ >= 50)
	{
		if (bit_is_clear(MFPGPDR, 0))
			bit_set(MFPGPDR, 0);
		else
			bit_clear(MFPGPDR, 0);
		c = 0;
	}
}

/*
 * Enable heartbeat LED
 */
void enable_hb_led()
{
	MFPTBCR = 0;
	MFPTBDR = 255;

	// enable interrupt for timer B
	bit_set(MFPIERA, 0);
	bit_set(MFPIMRA, 0);

	set_vector(72, timerb_isr);

	// start timer B
	MFPTBCR = 7;
}

/*
 * Disable heartbeat LED
 */
void disable_hb_led()
{
	// disable interrupt for timer B
	bit_clear(MFPIERA, 0);
	bit_clear(MFPIMRA, 0);

	// stop timer B
	MFPTBCR = 0;
}

/*
 * Force various exceptions
 */
void test_exception(int n)
{
	int a,b,c;
	uint8_t *dp;

	switch (n)
	{
		// bus error
		case 2:
			dp = (uint8_t *)0x1f0000;
			*dp = 0;
			break;
		// illegal instruction
		case 4:
			__asm("move.l #0,%a0");
			__asm("jsr @(%a0)");
			break;
		// divide by zero
		case 5:
			a = 1;
			b = 0;
			c = a / b;
			mfp_printf("%d", c);
			break;
		// trap 0
		case 32:
			__asm("trap #0");
			break;
		// trap 15
		case 47:
			__asm("trap #15");
			break;
		default:
			mfp_printf("Unimplemented exception\r\n");
	}
}

/*
 * Mandelbrot floating point test
 *
 * Source: http://archive.atomicmpc.com.au/forums.asp?s=2&c=10&t=2797
 */
void print_mandel()
{
	int a, b;
	double x, y;
	const int maxiter = 32;
	int iter;

	for (b = 0; b <= 32; b++) /* rows */
	{
		for (a = 0; a <= 85; a++) /* columns */
		{
			/* mandelbrot is an iteration of z1 = z^2 + c
			 here, zi and zr are the imaginary and real
			 portions of z */
			double zi = 0.0, zr = 0.0;

			/* and zni and znr are the imaginary and real
			 portions of z1 */
			double zni = 0.0, znr = 0.0;

			/* map ascii x/y coords to the actual
			 x/y values in the mandelbrot set */
			x = (a - 50) / 20.0;
			y = (b - 16.0) / 10.0;

			iter = 1;
			zi = 0.0;
			zr = 0.0;
			do
			{
				/* imaginary portion of z^2 + c */
				zni = 2 * zi * zr + y;
				/* real portion of z^2 + c */
				znr = zr * zr - zi * zi + x;
				/* new values for next iteration */
				zi = zni;
				zr = znr;
				/* is the distance from the origin
				 greater than 2?  if so, break
				 From memory, this condition means the
				 iteration will diverge to infinity, so
				 no point continuing */
				if (zi * zi + zr * zr > 4.0)
					break;

				iter++;
			} while (iter < maxiter);
			if (iter == maxiter)
				mfp_putc(' ');
			else
				mfp_putc(" .:-;!/>)|&IH%*#"[iter & 15]);
		}
		mfp_putc('\r');
		mfp_putc('\n');
	}
}

