/*
 * T030 ROM BIOS - SD card support
 *
 * (c) 2016 Tobias Rathje <tobbe@fasanix.com>
 *
 */

#include <stdio.h>
#include <string.h>
#include "tp3465.h"
#include "mfp.h"
#include "vector.h"
#include "misc.h"
#include "sd.h"
#include "lxload.h"
#include "netload.h"

#define DRV_MMC 1

#include "ff.h"
#include "diskio.h"
#include "mmc.h"

FATFS fs;
BYTE fbuf[4096];

/*
 * ISR for timer A
 */
void __attribute__ ((interrupt)) timera_isr()
{
	mmc_disk_timerproc();
}

/*
 * Initialize SD module
 */
void sd_init()
{
	// init SPI driver
	tp3465_init();

	// init timer interrupt
	MFPTADR = 0x80;
	MFPTACR = 0x07;
	bit_set(MFPIERA, 5);
	bit_set(MFPIMRA, 5);
	set_vector(77, timera_isr);
}

/*
 * Check fs operation result
 */
int chk_res(char *op, FRESULT res)
{
	if (res)
	{
		mfp_printf("%s error: %d\r\n", op, res);
		return 1;
	}
	else
		return 0;
}

/*
 * Print progress bar
 */
void print_prog_bar()
{
	mfp_printf("0%%");
	for (int i = 0; i < 41; i++)
		mfp_printf("_");
	mfp_printf("100%%\r\n");
}

/*
 * SD directory listing
 */
void sd_dir()
{
	char buf[60];
	long p1;
	UINT s1, s2;
	FATFS *ffs;
	DIR dir;
	FILINFO finfo;
	FRESULT res;
	DWORD ser;


	if (chk_res("f_mount", f_mount(&fs, "0:", 1)))
		return;

	if (chk_res("f_getlabel", f_getlabel("0:", buf, &ser)))
		return;

	mfp_printf("\r\nVolume label: %s\r\n\r\n", buf);

	char *path = "0:";

	if (chk_res("f_opendir", f_opendir(&dir, path)))
		return;

	p1 = s1 = s2 = 0;

	for (;;)
	{
		res = f_readdir(&dir, &finfo);
		if ((res != FR_OK) || !finfo.fname[0])
			break;
		if (finfo.fattrib & AM_DIR)
		{
			s2++;
		}
		else
		{
			s1++;
			p1 += finfo.fsize;
		}
		mfp_printf("%c%c%c%c%c %u/%02u/%02u %02u:%02u %9lu  %s\r\n",
				(finfo.fattrib & AM_DIR) ? 'D' : '-',
				(finfo.fattrib & AM_RDO) ? 'R' : '-',
				(finfo.fattrib & AM_HID) ? 'H' : '-',
				(finfo.fattrib & AM_SYS) ? 'S' : '-',
				(finfo.fattrib & AM_ARC) ? 'A' : '-', (finfo.fdate >> 9) + 1980,
				(finfo.fdate >> 5) & 15, finfo.fdate & 31, (finfo.ftime >> 11),
				(finfo.ftime >> 5) & 63, (DWORD) finfo.fsize, finfo.fname);
	}
	if (res == FR_OK)
	{
		mfp_printf("%4u File(s),%10lu bytes total\r\n%4u Dir(s)", s1, p1, s2);
		if (f_getfree(path, (DWORD*) &p1, &ffs) == FR_OK)
		{
			mfp_printf(", %10luKiB free\n", p1 * ffs->csize / 2);
		}
	}

	// unmount
	f_mount(NULL, "0:", 0);
}

/*
 * Read file from SD card to memory
 */
void sd_read_file(char *filename, uint32_t addr)
{
	FIL file;
	FSIZE_t fsize;
	unsigned int br;
	uint32_t pc = 0, bc = 0;
	uint32_t pbite, pmin;
	volatile unsigned char* mp;

	if (chk_res("f_mount", f_mount(&fs, "0:", 1)))
		return;

	if (chk_res("f_open", f_open(&file, filename, FA_READ)))
		return;

	fsize = f_size(&file);
	pbite = fsize / 50;
	pmin = sizeof fbuf * 50;

	if (fsize > pmin)
		print_prog_bar();

	while(1)
	{
		if (chk_res("f_read", f_read(&file, &fbuf, sizeof fbuf, &br)))
				break;

		bc += br;
		pc += br;

		for (int i = 0; i < br; i++)
		{
			mp = (unsigned char*)addr;
			*mp = fbuf[i];
			addr++;
		}

		if (fsize > pmin && pc > pbite)
		{
			mfp_printf("#");
			pc = 0;
		}

		if (br < sizeof fbuf)
			break;
	}

	mfp_printf("\r\n\r\n%d bytes read\r\n", bc);

	last_xfer_size = bc;

	if (chk_res("f_close", f_close(&file)))
			return;

	// unmount
	f_mount(NULL, "0:", 0);
}

/*
 * Write file from memory to SD card
 */
void sd_write_file(char *filename, uint32_t addr, uint32_t len)
{
	FIL file;
	unsigned int bw;
	uint32_t pc = 0, bc = 0;
	uint32_t pbite, pmin;
	uint16_t cc = 0;
	volatile unsigned char* mp;

	if (chk_res("f_mount", f_mount(&fs, "0:", 1)))
		return;

	if (chk_res("f_open", f_open(&file, filename, FA_CREATE_ALWAYS | FA_WRITE | FA_READ)))
		return;

	pbite = len / 50;
	pmin = sizeof fbuf * 50;

	if (len > pmin)
		print_prog_bar();

	while(1)
	{
		cc = 0;

		for (int i = 0; i < sizeof fbuf; i++)
		{
			mp = (unsigned char*)addr;
			fbuf[i] = *mp;
			addr++;
			bc++;
			cc++;
			pc++;

			if (bc == len)
				break;
		}

		if (chk_res("f_write", f_write(&file, &fbuf, cc, &bw)))
				break;

		if (len > pmin && pc > pbite)
		{
			mfp_printf("#");
			pc = 0;
		}

		if (bc == len)
			break;
	}

	mfp_printf("\r\n\r\n%d bytes written\r\n", bc);

	if (chk_res("f_close", f_close(&file)))
			return;

	// unmount
	f_mount(NULL, "0:", 0);
}

/*
 * Delete file on SD card
 */
void sd_delete_file(char *filename)
{
	if (chk_res("f_mount", f_mount(&fs, "0:", 1)))
		return;

	chk_res("f_unlink", f_unlink(filename));

	f_mount(NULL, "0:", 0);
}

/*
 * Check for init script (0)
 */
int sd_chk_init_script()
{
	FRESULT res;
	FILINFO finfo;

	res = f_mount(&fs, "0:", 1);

	if (res)
		return 0;

	res = f_stat("0.cmd", &finfo);

	f_mount(NULL, "0:", 0);

	if (res == FR_OK)
		return 1;
	else
		return 0;
}

/*
 * Load script from SD card into buffer
 */
int sd_load_script(int snum, char *sbuf)
{
	FIL file;
	FSIZE_t fsize;
	unsigned int br;
	char filename[20];

	if (chk_res("f_mount", f_mount(&fs, "0:", 1)))
		return -1;

	siprintf(filename, "%d.cmd", snum);

	if (chk_res("f_open", f_open(&file, filename, FA_READ)))
		return -2;

	fsize = f_size(&file);

	while(1)
	{
		if (chk_res("f_read", f_read(&file, &fbuf, sizeof fbuf, &br)))
				break;

		for (int i = 0; i < br; i++)
		{
			*sbuf++ = fbuf[i];
		}

		if (br < sizeof fbuf)
			break;
	}

	if (chk_res("f_close", f_close(&file)))
			return -1;

	// unmount
	f_mount(NULL, "0:", 0);

	return fsize;
}

/*
 * Check SD parameter
 */
int chk_param(int p, int ep)
{
	if (p != ep)
	{
		mfp_printf("Missing parameters [%d,%d]\r\n", p, ep);
		return 0;
	}
	else
		return 1;
}

/*
 * SD command interpreter
 */
void sd_cmd(char *param)
{
	char filename[60];
	uint32_t addr, len;

	char cmd = *param++;

	param++;

	switch (cmd)
	{
		case 'd':
			if (chk_param(siscanf(param, "%s", filename), 1))
				sd_delete_file(filename);
			break;
		case 'l':
			sd_dir();
			break;
		case 'n':
			if (chk_param(siscanf(param, "%s", filename), 1))
			{
				addr = 0x1000000;
				do_netload(LCL_IP_DHCP, DEFAULT_TFTP_SERVER, filename, addr);
				mfp_printf("Writing...\r\n\r\n");
				sd_write_file(filename, addr, last_xfer_size);
			}
			break;
		case 'r':
			if (chk_param(siscanf(param, "%s %"SCNx32, filename, &addr), 2))
				sd_read_file(filename, addr);
			break;
		case 'w':
			if (chk_param(siscanf(param, "%s %"SCNx32" %ld", filename, &addr, &len), 3))
				sd_write_file(filename, addr, len);
			break;
		default:
			mfp_printf("Unknown s cmd: %s\r\n", --param);
	}
}
